/*
* 07/08/2003 - 19:09:03
*
* XMLIDGenerator.java - 
* Original source: http://java.sun.com/xml/jaxp/dist/1.1/docs/tutorial/sax/work/Echo12.java
* Copyright (C) 2003 Ashwin Jayaprakash
* ashwinjay@rediffmail.com
* www.javaforu.com
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package searchassist.utils;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.HashMap;
import java.util.Map;
import java.util.Stack;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.ext.LexicalHandler;
import org.xml.sax.helpers.DefaultHandler;

/**
 * @author Ashwin Jayaprakash
 */
public class XMLIDGenerator extends DefaultHandler implements LexicalHandler
{
	private Writer out;
	private String indentString = "   "; // Amount to indent
	private int indentLevel = 0;
	private String lineEnd = System.getProperty("line.separator");

	private Stack level = new Stack();
	private Map mapOfIDForLevel = new HashMap();
	private StringBuffer strBuffer = new StringBuffer();

	private static final String BATON = "Baton";
	private static final boolean DEBUG = true;

	public XMLIDGenerator(Writer outWriter)
	{
		out = outWriter;
	} //end constr

	//===========================================================
	// SAX DocumentHandler methods
	//===========================================================

	public void setDocumentLocator(Locator l)
	{
		// Save this to resolve relative URIs or to give diagnostics.
		try
		{
			out.flush();
		}
		catch (IOException e)
		{
			// Ignore errors
		}
	}

	public void startDocument() throws SAXException
	{
		emit("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
		nl();
	}

	public void startElement(String namespaceURI, String lName, // local name
	String qName, // qualified name
	Attributes attrs) throws SAXException
	{
		indentLevel++;
		nl();
		String eName = lName; // element name
		if ("".equals(eName))
		{
			eName = qName; // namespaceAware = false
		} //end if

		//Clear the buffer
		strBuffer.delete(0, strBuffer.length());
		if (addUniqueIDForTag(eName))
		{
			level.push(BATON);

			Integer currIndentLevel = new Integer(level.size());
			//If no one else has there at this level before.
			if (mapOfIDForLevel.get(currIndentLevel) == null)
			{
				//start afresh
				mapOfIDForLevel.put(currIndentLevel, new Integer(0));
			} //end if

			//Linked list is zero based.
			Integer latestID = (Integer) mapOfIDForLevel.get(currIndentLevel);
			Integer currID = new Integer(latestID.intValue() + 1);
			//Put latest id number at this level
			mapOfIDForLevel.put(currIndentLevel, currID);

			int size = level.size();
			for (int i = 1; i <= size; i++)
			{
				Integer array_element = (Integer) mapOfIDForLevel.get(new Integer(i));
				strBuffer.append(array_element);
				if (i < size)
				{
					strBuffer.append('_');
				} //end inner if
			} //end for
		} //end if

		emit("<" + eName);
		if (attrs != null)
		{
			for (int i = 0; i < attrs.getLength(); i++)
			{
				String aName = attrs.getLocalName(i); // Attr name 
				if ("".equals(aName))
				{
					aName = attrs.getQName(i);
				} //end if
				emit("\t");
				emit(aName);
				emit("=\"");
				emit(attrs.getValue(i));
				emit("\"");
			} //end for
		} //end if

		if (strBuffer.length() > 0)
		{
			emit("\t");
			emit(getUniqueIDAttrName());
			emit("=\"");
			emit(strBuffer.toString());
			emit("\"");
		} //end if
		emit(">");
	}

	public void characters(char buf[], int offset, int len) throws SAXException
	{
		String s = new String(buf, offset, len).trim();
		if (!s.equals(""))
		{
			//			nl();
			emit(s);
		} //end if
	}

	public void ignorableWhitespace(char buf[], int offset, int len) throws SAXException
	{
		// Ignore it
	}

	public void processingInstruction(String target, String data) throws SAXException
	{
		nl();
		emit("<?" + target + " " + data + "?>");
	}

	public void endElement(String namespaceURI, String sName, // simple name
	String qName // qualified name
	) throws SAXException
	{
		String eName = sName; // element name
		if ("".equals(eName))
		{
			eName = qName; // namespaceAware = false
		} //end if

		if (addUniqueIDForTag(eName))
		{
			Integer currIndentLevel = new Integer(level.size());

			//clean up index for higher level - Unwind.
			mapOfIDForLevel.remove(new Integer(currIndentLevel.intValue() + 1));

			level.pop();
		} //end if

		nl();
		emit("</" + eName + ">");
		indentLevel--;
	}

	public void endDocument() throws SAXException
	{
		try
		{
			nl();
			out.flush();
		}
		catch (IOException e)
		{
			throw new SAXException("I/O error", e);
		}
	}

	//===========================================================
	// SAX ErrorHandler methods
	//===========================================================

	// treat validation errors as fatal
	public void error(SAXParseException e) throws SAXParseException
	{
		throw e;
	}

	// dump warnings too
	public void warning(SAXParseException err) throws SAXParseException
	{
		System.out.println("** Warning" + ", line " + err.getLineNumber() + ", uri " + err.getSystemId());
		System.out.println("   " + err.getMessage());
	}

	//===========================================================
	// LexicalEventListener methods
	//===========================================================

	public void comment(char[] ch, int start, int length) throws SAXException
	{
	}

	public void startCDATA() throws SAXException
	{
	}

	public void endCDATA() throws SAXException
	{
	}

	public void startEntity(java.lang.String name) throws SAXException
	{
	}

	public void endEntity(java.lang.String name) throws SAXException
	{
	}

	public void startDTD(String name, String publicId, String systemId) throws SAXException
	{
	}

	public void endDTD() throws SAXException
	{
	}

	protected boolean addUniqueIDForTag(String tagName)
	{
		boolean retValue = false;
		if (tagName.equals("category") || tagName.equals("item"))
		{
			retValue = true;
		} //end if

		return retValue;
	} //end method

	protected String getUniqueIDAttrName()
	{
		return "uniqueID";
	} //end method

	//===========================================================
	// Utility Methods ...
	//===========================================================

	// Wrap I/O exceptions in SAX exceptions, to
	// suit handler signature requirements
	private void emit(String s) throws SAXException
	{
		try
		{
			out.write(s);
			out.flush();
		}
		catch (IOException e)
		{
			throw new SAXException("I/O error", e);
		}
	}

	// Start a new line
	// and indent the next line appropriately
	private void nl() throws SAXException
	{
		try
		{
			out.write(lineEnd);
			for (int i = 0; i < indentLevel; i++)
				out.write(indentString);
		}
		catch (IOException e)
		{
			throw new SAXException("I/O error", e);
		}
	}

	public static void main(String argv[])
	{
		if (argv.length < 2)
		{
			System.err.println("Usage: cmd input_filename output_filename");
			System.exit(1);
		}

		try
		{
			FileOutputStream fos = new FileOutputStream(new File(argv[1]));
			// Set up output stream
			//                        Writer pw = new OutputStreamWriter(System.out, "UTF8");                      
			Writer pw = new OutputStreamWriter(new BufferedOutputStream(fos), "UTF8");
			// Use an instance of ourselves as the SAX event handler
			DefaultHandler handler = new XMLIDGenerator(pw);

			// Use the validating parser
			SAXParserFactory factory = SAXParserFactory.newInstance();
			factory.setValidating(false);
			//                        factory.setNamespaceAware(true);

			// Parse the input
			SAXParser saxParser = factory.newSAXParser();
			XMLReader xmlReader = saxParser.getXMLReader();
			xmlReader.setProperty("http://xml.org/sax/properties/lexical-handler", handler);
			saxParser.parse(new File(argv[0]), handler);
		}
		catch (SAXParseException spe)
		{
			// Error generated by the parser
			System.out.println("\n** Parsing error" + ", line " + spe.getLineNumber() + ", uri " + spe.getSystemId());
			System.out.println("   " + spe.getMessage());

			// Use the contained exception, if any
			Exception x = spe;
			if (spe.getException() != null)
			{
				x = spe.getException();
			} //end if
			x.printStackTrace();
		}
		catch (SAXException sxe)
		{
			// Error generated by this application
			// (or a parser-initialization error)
			Exception x = sxe;
			if (sxe.getException() != null)
			{
				x = sxe.getException();
			} //end if
			x.printStackTrace();
		}
		catch (ParserConfigurationException pce)
		{
			// Parser with specified options can't be built
			pce.printStackTrace();
		}
		catch (IOException ioe)
		{
			// I/O error
			ioe.printStackTrace();
		}

		System.exit(0);
	} //end main

} //end class
